@@ -18,8 +18,10 @@ This module exports useful functions for 2/3 compatible code:
     * types:
 
         * text_type: unicode in Python 2, str in Python 3
-        * binary_type: str in Python 2, bythes in Python 3
         * string_types: basestring in Python 2, str in Python 3
+        * binary_type: str in Python 2, bytes in Python 3
+        * integer_types: (int, long) in Python 2, int in Python 3
+        * class_types: (type, types.ClassType) in Python 2, type in Python 3
 
     * bchr(c):
         Take an integer and make a 1-character byte string
@@ -55,7 +57,10 @@ import copy
 import inspect
 
 
-PY3 = sys.version_info[0] == 3
+PY3 = sys.version_info[0] >= 3
+PY34_PLUS = sys.version_info[0:2] >= (3, 4)
+PY35_PLUS = sys.version_info[0:2] >= (3, 5)
+PY36_PLUS = sys.version_info[0:2] >= (3, 6)
 PY2 = sys.version_info[0] == 2
 PY26 = sys.version_info[0:2] == (2, 6)
 PY27 = sys.version_info[0:2] == (2, 7)
@@ -403,12 +408,34 @@ if PY3:
         allows re-raising exceptions with the cls value and traceback on
         Python 2 and 3.
         """
-        if value is not None and isinstance(tp, Exception):
-            raise TypeError("instance exception may not have a separate value")
-        if value is not None:
-            exc = tp(value)
-        else:
+        if isinstance(tp, Exception):
+            # If the first object is an instance, the type of the exception
+            # is the class of the instance, the instance itself is the value,
+            # and the second object must be None.
+            if value is not None:
+                raise TypeError("instance exception may not have a separate value")
             exc = tp
+        elif not issubclass(tp, Exception):
+            # If the first object is a class, it becomes the type of the
+            # exception.
+            raise TypeError("class must derive from Exception")
+        else:
+            # The second object is used to determine the exception value: If it
+            # is an instance of the class, the instance becomes the exception
+            # value. If the second object is a tuple, it is used as the argument
+            # list for the class constructor; if it is None, an empty argument
+            # list is used, and any other object is treated as a single argument
+            # to the constructor. The instance so created by calling the
+            # constructor is used as the exception value.
+            if isinstance(value, tp):
+                exc = value
+            elif isinstance(value, tuple):
+                exc = tp(*value)
+            elif value is None:
+                exc = tp()
+            else:
+                exc = tp(value)
+
         if exc.__traceback__ is not tb:
             raise exc.with_traceback(tb)
         raise exc
@@ -441,12 +468,14 @@ else:
         e.__suppress_context__ = False
         if isinstance(cause, type) and issubclass(cause, Exception):
             e.__cause__ = cause()
+            e.__cause__.__traceback__ = sys.exc_info()[2]
             e.__suppress_context__ = True
         elif cause is None:
             e.__cause__ = None
             e.__suppress_context__ = True
         elif isinstance(cause, BaseException):
             e.__cause__ = cause
+            e.__cause__.__traceback__ = sys.exc_info()[2]
             e.__suppress_context__ = True
         else:
             raise TypeError("exception causes must derive from BaseException")
@@ -550,15 +579,14 @@ def isbytes(obj):
 
 def isnewbytes(obj):
     """
-    Equivalent to the result of ``isinstance(obj, newbytes)`` were
-    ``__instancecheck__`` not overridden on the newbytes subclass. In
-    other words, it is REALLY a newbytes instance, not a Py2 native str
+    Equivalent to the result of ``type(obj)  == type(newbytes)``
+    in other words, it is REALLY a newbytes instance, not a Py2 native str
     object?
+
+    Note that this does not cover subclasses of newbytes, and it is not
+    equivalent to ininstance(obj, newbytes)
     """
-    # TODO: generalize this so that it works with subclasses of newbytes
-    # Import is here to avoid circular imports:
-    from future.types.newbytes import newbytes
-    return type(obj) == newbytes
+    return type(obj).__name__ == 'newbytes'
 
 
 def isint(obj):
@@ -724,16 +752,16 @@ else:
 
 
 __all__ = ['PY2', 'PY26', 'PY3', 'PYPY',
-           'as_native_str', 'bind_method', 'bord', 'bstr',
-           'bytes_to_native_str', 'encode_filename', 'ensure_new_type',
-           'exec_', 'get_next', 'getexception', 'implements_iterator',
-           'is_new_style', 'isbytes', 'isidentifier', 'isint',
-           'isnewbytes', 'istext', 'iteritems', 'iterkeys', 'itervalues',
-           'lfilter', 'listitems', 'listvalues', 'lmap', 'lrange',
-           'lzip', 'native', 'native_bytes', 'native_str',
+           'as_native_str', 'binary_type', 'bind_method', 'bord', 'bstr',
+           'bytes_to_native_str', 'class_types', 'encode_filename',
+           'ensure_new_type', 'exec_', 'get_next', 'getexception',
+           'implements_iterator', 'integer_types', 'is_new_style', 'isbytes',
+           'isidentifier', 'isint', 'isnewbytes', 'istext', 'iteritems',
+           'iterkeys', 'itervalues', 'lfilter', 'listitems', 'listvalues',
+           'lmap', 'lrange', 'lzip', 'native', 'native_bytes', 'native_str',
            'native_str_to_bytes', 'old_div',
            'python_2_unicode_compatible', 'raise_',
-           'raise_with_traceback', 'reraise', 'text_to_native_str',
-           'tobytes', 'viewitems', 'viewkeys', 'viewvalues',
-           'with_metaclass'
-          ]
+           'raise_with_traceback', 'reraise', 'string_types',
+           'text_to_native_str', 'text_type', 'tobytes', 'viewitems',
+           'viewkeys', 'viewvalues', 'with_metaclass'
+           ]
